#include "precompile.h"

#include <string.h>

#include <a8/udplog.h>

#include "navmeshbuilder.h"
#include "mapinstance.h"
#include "collider.h"
#include "entity.h"
#include "metamgr.h"

#include "navmeshhelper.h"

#include "framework/cpp/inputgeom.h"

void NavMeshBuilder::Init()
{

}

void NavMeshBuilder::UnInit()
{

}

void NavMeshBuilder::Build(MapInstance* map_instance)
{
    BuilderParams builder_params;
    {
        builder_params.map_instance = map_instance;
        builder_params.geom = new f8::InputGeom();
        builder_params.geom->Init(map_instance->map_meta_->i->map_width(),
                                  map_instance->map_meta_->i->map_height());
        builder_params.ctx = new BuildContext();
    }
    InitBuilderParams(builder_params);
    CreateTileCache(builder_params);
    CreateNavMesh(builder_params);
    BuildTiles(builder_params);
    #if 0
    BuildMapObstalce(builder_params);
    #endif
    map_instance->navmesh_ = builder_params.navmesh;
    Save(builder_params);
}

void NavMeshBuilder::InitBuilderParams(BuilderParams& builder_params)
{
    // Init cache
    rcCalcGridSize(builder_params.geom->GetMeshBoundsMin(),
                   builder_params.geom->GetMeshBoundsMax(),
                   builder_params.kCellSize,
                   &builder_params.grid_width,
                   &builder_params.grid_height);
    builder_params.tile_size = (int)builder_params.kTileSize;
    builder_params.tile_width = (builder_params.grid_width + builder_params.tile_size-1) /
        builder_params.tile_size;
    builder_params.tile_height = (builder_params.grid_height + builder_params.tile_size-1) /
        builder_params.tile_size;
}

void NavMeshBuilder::InitTileCacheParams(BuilderParams& builder_params, dtTileCacheParams& tcparams)
{
    // Tile cache params.
    memset(&tcparams, 0, sizeof(tcparams));
    rcVcopy(tcparams.orig, builder_params.geom->GetMeshBoundsMin());
    tcparams.cs = builder_params.kCellSize;
    tcparams.ch = builder_params.kCellHeight;
    tcparams.width = (int)builder_params.kTileSize;
    tcparams.height = (int)builder_params.kTileSize;
    tcparams.walkableHeight = builder_params.kAgentHeight;
    tcparams.walkableRadius = builder_params.kAgentRadius;
    tcparams.walkableClimb = builder_params.kAgentMaxClimb;
    tcparams.maxSimplificationError = builder_params.kEdgeMaxError;
    tcparams.maxTiles = builder_params.tile_width * builder_params.tile_height * EXPECTED_LAYERS_PER_TILE;
    tcparams.maxObstacles = 128 * 1000;
}

void NavMeshBuilder::InitNavMeshParams(BuilderParams& builder_params, dtNavMeshParams& params)
{
    memset(&params, 0, sizeof(params));
    rcVcopy(params.orig, builder_params.geom->GetMeshBoundsMin());
    params.tileWidth = builder_params.kTileSize * builder_params.kCellSize;
    params.tileHeight = builder_params.kTileSize * builder_params.kCellSize;
    #if 1
    int tile_bits = rcMin((int)dtIlog2(
                                       dtNextPow2(params.tileWidth*params.tileHeight*EXPECTED_LAYERS_PER_TILE)),
                          14);
    if (tile_bits > 14) {
        tile_bits = 14;
    }
    int poly_bits = 22 - tile_bits;
    params.maxTiles = 1 << tile_bits;
    params.maxPolys = 1 << poly_bits;
    #else
    params.maxTiles = builder_params.kMaxTiles;
    params.maxPolys = builder_params.kMaxPolysPerTile;
    #endif
}

void NavMeshBuilder::BuildTiles(BuilderParams& builder_params)
{
    rcConfig cfg;
    {
        memset(&cfg, 0, sizeof(cfg));
        cfg.cs = builder_params.kCellSize;
        cfg.ch = builder_params.kCellHeight;
        cfg.walkableSlopeAngle = builder_params.kAgentMaxSlope;
        cfg.walkableHeight = (int)ceilf(builder_params.kAgentHeight / cfg.ch);
        cfg.walkableClimb = (int)floorf(builder_params.kAgentMaxClimb / cfg.ch);
        cfg.walkableRadius = (int)ceilf(builder_params.kAgentRadius / cfg.cs);
        cfg.maxEdgeLen = (int)(builder_params.kEdgeMaxLen / builder_params.kCellSize);
        cfg.maxSimplificationError = builder_params.kEdgeMaxError;
        cfg.minRegionArea = (int)rcSqr(builder_params.kRegionMinSize);		// Note: area = size*size
        cfg.mergeRegionArea = (int)rcSqr(builder_params.kRegionMergeSize);	// Note: area = size*size
        cfg.maxVertsPerPoly = (int)builder_params.kVertsPerPoly;
        cfg.tileSize = (int)builder_params.kTileSize;
        cfg.borderSize = cfg.walkableRadius + 3; // Reserve enough padding.
        cfg.width = cfg.tileSize + cfg.borderSize*2;
        cfg.height = cfg.tileSize + cfg.borderSize*2;
        cfg.detailSampleDist = builder_params.kDetailSampleDist < 0.9f ? 0 : builder_params.kCellSize * builder_params.kDetailSampleDist;
        cfg.detailSampleMaxError = builder_params.kCellHeight * builder_params.kDetailSampleMaxError;
        rcVcopy(cfg.bmin, builder_params.geom->GetMeshBoundsMin());
        rcVcopy(cfg.bmax, builder_params.geom->GetMeshBoundsMax());
    }

    for (int y = 0; y < builder_params.tile_height; ++y) {
        for (int x = 0; x < builder_params.tile_width; ++x) {
            TileCacheData tiles[MAX_LAYERS];
            memset(tiles, 0, sizeof(tiles));
            int ntiles = RasterizeTileLayers(builder_params, x, y, cfg, tiles, MAX_LAYERS);

            for (int i = 0; i < ntiles; ++i) {
                TileCacheData* tile = &tiles[i];
                dtStatus status = builder_params.tile_cache->addTile(tile->data,
                                                                     tile->dataSize,
                                                                     DT_COMPRESSEDTILE_FREE_DATA,
                                                                     0);
                if (dtStatusFailed(status)) {
                    abort();
                    dtFree(tile->data);
                    tile->data = 0;
                    continue;
                }
            }
        }
    }

    // Build initial meshes
    for (int y = 0; y < builder_params.tile_height; ++y) {
        for (int x = 0; x < builder_params.tile_width; ++x) {
            builder_params.tile_cache->buildNavMeshTilesAt(x, y, builder_params.navmesh);
        }
    }
}

int NavMeshBuilder::RasterizeTileLayers(BuilderParams& builder_params,
                                        const int tx,
                                        const int ty,
                                        const rcConfig& cfg,
                                        TileCacheData* tiles,
                                        const int maxTiles)
{
    rcAreaModification SAMPLE_AREAMOD_GROUND(SAMPLE_POLYAREA_TYPE_GROUND, SAMPLE_POLYAREA_TYPE_MASK);

    FastLZCompressor comp;
    RasterizationContext rc;

    const float* verts = builder_params.geom->GetVerts();
    const int nverts = builder_params.geom->GetVertCount();
    const rcChunkyTriMesh* chunkyMesh = builder_params.geom->GetChunkyMesh();

    // Tile bounds.
    const float tcs = cfg.tileSize * cfg.cs;

    rcConfig tcfg;
    memcpy(&tcfg, &cfg, sizeof(tcfg));

    tcfg.bmin[0] = cfg.bmin[0] + tx*tcs;
    tcfg.bmin[1] = cfg.bmin[1];
    tcfg.bmin[2] = cfg.bmin[2] + ty*tcs;
    tcfg.bmax[0] = cfg.bmin[0] + (tx+1)*tcs;
    tcfg.bmax[1] = cfg.bmax[1];
    tcfg.bmax[2] = cfg.bmin[2] + (ty+1)*tcs;
    tcfg.bmin[0] -= tcfg.borderSize*tcfg.cs;
    tcfg.bmin[2] -= tcfg.borderSize*tcfg.cs;
    tcfg.bmax[0] += tcfg.borderSize*tcfg.cs;
    tcfg.bmax[2] += tcfg.borderSize*tcfg.cs;

    // Allocate voxel heightfield where we rasterize our input data to.
    rc.solid = rcAllocHeightfield();
    if (!rc.solid) {
        return 0;
    }
    if (!rcCreateHeightfield(builder_params.ctx,
                             *rc.solid,
                             tcfg.width,
                             tcfg.height,
                             tcfg.bmin,
                             tcfg.bmax,
                             tcfg.cs,
                             tcfg.ch)) {
        return 0;
    }

    // Allocate array that can hold triangle flags.
    // If you have multiple meshes you need to process, allocate
    // and array which can hold the max number of triangles you need to process.
    rc.triareas = new unsigned char[chunkyMesh->maxTrisPerChunk];
    if (!rc.triareas) {
        builder_params.ctx->log(RC_LOG_ERROR,
                                "buildNavigation: Out of memory 'm_triareas' (%d).",
                                chunkyMesh->maxTrisPerChunk);
        return 0;
    }

    float tbmin[2], tbmax[2];
    tbmin[0] = tcfg.bmin[0];
    tbmin[1] = tcfg.bmin[2];
    tbmax[0] = tcfg.bmax[0];
    tbmax[1] = tcfg.bmax[2];
    int cid[512];// TODO: Make grow when returning too many items.
    const int ncid = rcGetChunksOverlappingRect(chunkyMesh, tbmin, tbmax, cid, 512);
    if (!ncid) {
        return 0; // empty
    }

    for (int i = 0; i < ncid; ++i) {
        const rcChunkyTriMeshNode& node = chunkyMesh->nodes[cid[i]];
        const int* tris = &chunkyMesh->tris[node.i*3];
        const int ntris = node.n;

        memset(rc.triareas, 0, ntris*sizeof(unsigned char));
        rcMarkWalkableTriangles(builder_params.ctx, tcfg.walkableSlopeAngle,
                                verts, nverts, tris, ntris, rc.triareas,
                                SAMPLE_AREAMOD_GROUND);

        if (!rcRasterizeTriangles(builder_params.ctx,
                                  verts,
                                  nverts,
                                  tris,
                                  rc.triareas,
                                  ntris,
                                  *rc.solid,
                                  tcfg.walkableClimb)) {
            return 0;
        }
    }

    // Once all geometry is rasterized, we do initial pass of filtering to
    // remove unwanted overhangs caused by the conservative rasterization
    // as well as filter spans where the character cannot possibly stand.
    if (builder_params.kFilterLowHangingObstacles) {
        rcFilterLowHangingWalkableObstacles(builder_params.ctx, tcfg.walkableClimb, *rc.solid);
    }
    if (builder_params.kFilterLedgeSpans) {
        rcFilterLedgeSpans(builder_params.ctx, tcfg.walkableHeight, tcfg.walkableClimb, *rc.solid);
    }
    if (builder_params.kFilterWalkableLowHeightSpans) {
        rcFilterWalkableLowHeightSpans(builder_params.ctx, tcfg.walkableHeight, *rc.solid);
    }

    rc.chf = rcAllocCompactHeightfield();
    if (!rc.chf) {
        builder_params.ctx->log(RC_LOG_ERROR, "buildNavigation: Out of memory 'chf'.");
        return 0;
    }
    if (!rcBuildCompactHeightfield(builder_params.ctx,
                                   tcfg.walkableHeight,
                                   tcfg.walkableClimb,
                                   *rc.solid,
                                   *rc.chf)) {
        builder_params.ctx->log(RC_LOG_ERROR, "buildNavigation: Could not build compact data.");
        return 0;
    }

    // Erode the walkable area by agent radius.
    if (!rcErodeWalkableArea(builder_params.ctx, tcfg.walkableRadius, *rc.chf)) {
        builder_params.ctx->log(RC_LOG_ERROR, "buildNavigation: Could not erode.");
        return 0;
    }

    #if 0
    // (Optional) Mark areas.
    const ConvexVolume* vols = builder_params.geom->GetConvexVolumes();
    for (int i  = 0; i < builder_params.geom->GetConvexVolumeCount(); ++i) {
        rcMarkConvexPolyArea(builder_params.ctx, vols[i].verts, vols[i].nverts,
                             vols[i].hmin, vols[i].hmax,
                             vols[i].areaMod, *rc.chf);
    }
    #endif

    rc.lset = rcAllocHeightfieldLayerSet();
    if (!rc.lset) {
        builder_params.ctx->log(RC_LOG_ERROR, "buildNavigation: Out of memory 'lset'.");
        return 0;
    }
    if (!rcBuildHeightfieldLayers(builder_params.ctx,
                                  *rc.chf,
                                  tcfg.borderSize,
                                  tcfg.walkableHeight,
                                  *rc.lset)) {
        builder_params.ctx->log(RC_LOG_ERROR, "buildNavigation: Could not build heighfield layers.");
        return 0;
    }

    rc.ntiles = 0;
    for (int i = 0; i < rcMin(rc.lset->nlayers, MAX_LAYERS); ++i) {
        TileCacheData* tile = &rc.tiles[rc.ntiles++];
        const rcHeightfieldLayer* layer = &rc.lset->layers[i];

        // Store header
        dtTileCacheLayerHeader header;
        header.magic = DT_TILECACHE_MAGIC;
        header.version = DT_TILECACHE_VERSION;

        // Tile layer location in the navmesh.
        header.tx = tx;
        header.ty = ty;
        header.tlayer = i;
        dtVcopy(header.bmin, layer->bmin);
        dtVcopy(header.bmax, layer->bmax);

        // Tile info.
        header.width = (unsigned char)layer->width;
        header.height = (unsigned char)layer->height;
        header.minx = (unsigned char)layer->minx;
        header.maxx = (unsigned char)layer->maxx;
        header.miny = (unsigned char)layer->miny;
        header.maxy = (unsigned char)layer->maxy;
        header.hmin = (unsigned short)layer->hmin;
        header.hmax = (unsigned short)layer->hmax;

        dtStatus status = dtBuildTileCacheLayer(&comp,
                                                &header,
                                                layer->heights,
                                                layer->areas,
                                                layer->cons,
                                                &tile->data,
                                                &tile->dataSize);
        if (dtStatusFailed(status)) {
            return 0;
        }
    }

    // Transfer ownsership of tile data from build context to the caller.
    int n = 0;
    for (int i = 0; i < rcMin(rc.ntiles, maxTiles); ++i) {
        tiles[n++] = rc.tiles[i];
        rc.tiles[i].data = 0;
        rc.tiles[i].dataSize = 0;
    }

    return n;
}

bool NavMeshBuilder::CreateTileCache(BuilderParams& builder_params)
{
    LinearAllocator* talloc = new LinearAllocator(320000);
    FastLZCompressor* tcomp = new FastLZCompressor();
    MeshProcess* tmproc = new MeshProcess();

    dtTileCacheParams tcparams;
    InitTileCacheParams(builder_params, tcparams);
    builder_params.tile_cache = dtAllocTileCache();
    dtStatus status = builder_params.tile_cache->init
        (&tcparams,
         talloc,
         tcomp,
         tmproc);
    if (dtStatusFailed(status)) {
        abort();
    }
    return true;
}

bool NavMeshBuilder::CreateNavMesh(BuilderParams& builder_params)
{
    dtNavMeshParams params;
    InitNavMeshParams(builder_params, params);

    builder_params.navmesh = dtAllocNavMesh();
    dtStatus status = builder_params.navmesh->init(&params);
    if (dtStatusFailed(status)) {
        abort();
    }
    return true;
}

void NavMeshBuilder::BuildMapObstalce(BuilderParams& builder_params)
{
    dtTileCache* tile_cache = builder_params.tile_cache;
    for (auto& pair : builder_params.map_instance->uniid_hash_) {
        for (ColliderComponent* collider : *pair.second->GetColliders()) {
            switch (collider->type) {
            case CT_Aabb:
                {
                    AabbCollider* aabb = (AabbCollider*)collider;

                    float bmin[3] = {0};
                    bmin[0] = (aabb->owner->GetPos() + aabb->_min).x;
                    bmin[1] = 0;
                    bmin[2] = (aabb->owner->GetPos() + aabb->_min).y;

                    float bmax[3] = {0};
                    bmax[0] = (aabb->owner->GetPos() + aabb->_max).x;
                    bmax[1] = 0;
                    bmax[2] = (aabb->owner->GetPos() + aabb->_max).y;

                    dtObstacleRef ref;
                    dtStatus status = tile_cache->addBoxObstacle(bmin,
                                                                 bmax,
                                                                 &ref
                                                                 );
                    if (status != DT_SUCCESS) {
                        abort();
                    }
                    {
                        bool uptodate = false;
                        int update_count = 0;
                        while (!uptodate) {
                            tile_cache->update(0, builder_params.navmesh, &uptodate);
                            ++update_count;
                            if (update_count > 10000 * 10) {
                                abort();
                            }
                        }
                    }
                }
                break;
            case CT_Circle:
                {
                    CircleCollider* circle = (CircleCollider*)collider;

                    float pos[3] = {0};
                    pos[0] = (circle->owner->GetPos() + circle->pos).x;
                    pos[1] = 0;
                    pos[2] = (circle->owner->GetPos() + circle->pos).y;

                    dtObstacleRef ref;
                    dtStatus status = tile_cache->addObstacle(pos,
                                                              circle->rad,
                                                              0,
                                                              &ref
                                                              );
                    if (status != DT_SUCCESS) {
                        abort();
                    }
                    {
                        bool uptodate = false;
                        int update_count = 0;
                        while (!uptodate) {
                            tile_cache->update(0, builder_params.navmesh, &uptodate);
                            ++update_count;
                            if (update_count > 10000 * 10) {
                                abort();
                            }
                        }
                    }
                }
                break;
            default:
                {
                }
                break;
            }
        }
    }//end for pair
}

void NavMeshBuilder::Save(BuilderParams& builder_params)
{
    FILE* fp = fopen("all_tiles_tilecache.bin", "wb");
    if (!fp) {
        abort();
    }

    static const int TILECACHESET_MAGIC = 'T'<<24 | 'S'<<16 | 'E'<<8 | 'T'; //'TSET';
    static const int TILECACHESET_VERSION = 1;

    struct TileCacheSetHeader
    {
	int magic;
	int version;
	int numTiles;
	dtNavMeshParams meshParams;
	dtTileCacheParams cacheParams;
    };

    struct TileCacheTileHeader
    {
	dtCompressedTileRef tileRef;
	int dataSize;
    };

    TileCacheSetHeader header;
    header.magic = TILECACHESET_MAGIC;
    header.version = TILECACHESET_VERSION;
    header.numTiles = 0;
    for (int i = 0; i < builder_params.tile_cache->getTileCount(); ++i) {
        const dtCompressedTile* tile = builder_params.tile_cache->getTile(i);
        if (!tile || !tile->header || !tile->dataSize) continue;
        header.numTiles++;
    }
    memcpy(&header.cacheParams, builder_params.tile_cache->getParams(), sizeof(dtTileCacheParams));
    memcpy(&header.meshParams, builder_params.navmesh->getParams(), sizeof(dtNavMeshParams));
    fwrite(&header, sizeof(TileCacheSetHeader), 1, fp);

    // Store tiles.
    for (int i = 0; i < builder_params.tile_cache->getTileCount(); ++i) {
        const dtCompressedTile* tile = builder_params.tile_cache->getTile(i);
        if (!tile || !tile->header || !tile->dataSize) continue;

        TileCacheTileHeader tileHeader;
        tileHeader.tileRef = builder_params.tile_cache->getTileRef(tile);
        tileHeader.dataSize = tile->dataSize;
        fwrite(&tileHeader, sizeof(tileHeader), 1, fp);

        fwrite(tile->data, tile->dataSize, 1, fp);
    }

    fclose(fp);
    a8::UdpLog::Instance()->Debug("Save OK", {});
}
